/****************************************************************************/

#include "traydebug.h"

/****************************************************************************/

// This is the class which handles and dispatches messages for the main window
// in the application.
class CMainWndDispatcher : public CDispatcher {
 public:
     CMainWndDispatcher (handlerType handler) : CDispatcher (handler) {}

     LRESULT MsgDestroy    (HWND , UINT , WPARAM , LPARAM);
     LRESULT MsgSize       (HWND , UINT , WPARAM , LPARAM);
     LRESULT MsgTrayNotify (HWND , UINT , WPARAM , LPARAM);
     LRESULT MsgSetFocus   (HWND , UINT , WPARAM , LPARAM);
     LRESULT MsgMenuEnable (HWND , UINT , WPARAM , LPARAM);

     LRESULT MsgOutputDebugMsg (HWND , UINT , WPARAM , LPARAM);

     LRESULT CmdToggleWindow (HWND , WORD , WORD , HWND);
     LRESULT CmdClearDebug   (HWND , WORD , WORD , HWND);
     LRESULT CmdExit         (HWND , WORD , WORD , HWND);

 MESSAGE_LIST;
 COMMAND_LIST;
};

START_MESSAGE_LIST (CMainWndDispatcher)
    MESSAGE (WM_TRAYNOTIFY      , MsgTrayNotify)
    MESSAGE (WM_WINLIBDEBUGMSG  , MsgOutputDebugMsg)
    MESSAGE (WM_SETFOCUS        , MsgSetFocus)
    MESSAGE (WM_INITMENUPOPUP   , MsgMenuEnable)
    MESSAGE (WM_SIZE            , MsgSize)
    MESSAGE (WM_DESTROY         , MsgDestroy)
END_MESSAGE_LIST

START_COMMAND_LIST (CMainWndDispatcher)
    COMMAND (IDM_DISPLAYDEBUG , CmdToggleWindow)
    COMMAND (IDM_CLEARDEBUG   , CmdClearDebug)
    COMMAND (IDM_EXITPROGRAM  , CmdExit)
END_COMMAND_LIST

/****************************************************************************/

CMainWndDispatcher MainWindowDispatch (handlerWindow);

/*****************************************************************************
 * Name         - WndProc
 * Purpose      - This handles messages sent to the main window. It uses the
 *                windows dispatching class to do it's job.
 * Arguments    - Standard window procedure parameters
 * Return Value - Depends on the message processes
 ****************************************************************************/
LRESULT CALLBACK WndProc (HWND hWnd , UINT uMessage,
                          WPARAM wParam , LPARAM lParam)
{
 return MainWindowDispatch.dispatchMessage (hWnd, uMessage, wParam, lParam);
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgSetFocus
 * Purpose      - This handles the WM_SETFOCUS message sent to the windows.
 *                If the window is the main window it is ignored because the
 *                main window is hidden. Otherwise, the edit control that is
 *                in the window is given the input focus
 * Arguments    - hWnd: Handle of the window getting the message
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgSetFocus (HWND hWnd , UINT /*uMessage*/,
                                         WPARAM /*wParam*/ , LPARAM /*lParam*/)
{
 if (hWnd == hwndDebugClient)
     SetFocus (hwndDebugEdit);

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgSize
 * Purpose      - This handles the WM_SIZEmessage sent to the windows.
 *                If the window is the main window it is ignored because the
 *                main window is hidden. Otherwise, the edit control that is
 *                in the window is resized to fill the entire window.
 * Arguments    - hWnd: Handle of the window getting the message
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgSize (HWND hWnd , UINT /*uMessage*/,
                                     WPARAM /*wParam*/ , LPARAM /*lParam*/)
{
 RECT Client;

 // Is the window that changed size the debug client window?
 if (hWnd == hwndDebugClient) {
     // Determine the client area of the window
     GetClientRect (hWnd , &Client);

     // Resize the debug edit control to fill the window.
     MoveWindow (hwndDebugEdit , 0 , 0 ,
                 (Client.right - Client.left + 1) ,
                 (Client.bottom - Client.top + 1) ,
                 TRUE);
 }

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgDestroy
 * Purpose      - This handles the WM_DESTROY message sent to the window. It
 *                posts a quit message to the main window, which causes the
 *                message loop to terminte and make the application terminate.
 * Arguments    - hWnd: Handle of the window getting the message
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgDestroy (HWND hWnd , UINT /*uMessage*/,
                                        WPARAM /*wParam*/ , LPARAM /*lParam*/)
{
 // If this is the main window being destroyed, then we want to remove the
 // tray icon and post a quit message to make the program terminate
 if (hWnd == hwndMain) {
     // Set up the structure that tells windows what icon to nuke
     NOTIFYICONDATA nid;

     nid.cbSize = sizeof (NOTIFYICONDATA);
     nid.hWnd = hWnd;
     nid.uID = 0;
     nid.uFlags = 0;

     // Delete the icon
     Shell_NotifyIcon (NIM_DELETE , &nid);

     // If there is a debug client or a note client tell them to close.
     if (hwndDebugClient)
         PostMessage (hwndDebugClient , WM_CLOSE , 0 , 0);

     // Post a quit message to shut things down
     PostQuitMessage (0);
 }

 // If this is the debug client window, then shift ownership of the debug
 // edit window back to the parent before the window closes and set the global
 // handle to NULL so we know the window doesn't exist.
 if (hWnd == hwndDebugClient) {
     // Before we tell the window to close, save it's position into the
     // registry.
     SaveDebugWindowPosition ();

     // Mark the debug client handle as nonexistant so the rest of the program
     // knows that the window is gone
     hwndDebugClient = NULL;

     // Change the parent of the debug edit control to the main window so that
     // it doesn't get lost.
     SetParent (hwndDebugEdit , hwndMain);
 }

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgOutputDebugMsg
 * Purpose      - This handles the WM_WINNOTEDEBUGMSG message from
 *                WinLib. It adds the message given to it to the debug
 *                edit control.
 * Arguments    - lParam: Global atom containing the text
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgOutputDebugMsg (HWND /*hWnd*/ ,
                                               UINT /*uMessage*/,
                                               WPARAM /*wParam*/ ,
                                               LPARAM lParam)
{
 // Add this string to the edit control. First, move the caret to the end of
 // the edit control.
 SendMessage (hwndDebugEdit , EM_SETSEL , -1 , -1);

 // Now add the text
 char Buffer[300];
 GlobalGetAtomName ((ATOM) lParam , Buffer , sizeof (Buffer));
 SendMessage (hwndDebugEdit , EM_REPLACESEL , 0 , (LPARAM) Buffer);

 // Add a newline
 SendMessage (hwndDebugEdit , EM_REPLACESEL , 0 , (LPARAM) "\r\n");

 // Make sure the caret is in view
 SendMessage (hwndDebugEdit , EM_SCROLLCARET , 0 , 0);

 return 0;
}

/*****************************************************************************
 * Name         - AlterMenuText
 * Purpose      - This function alters the menu text of a particular menu
 *                item (based on it's command ID) to the given text.
 * Arguments    - hMenu: Handle to the menu containing the item to alter
 *                ID: Command ID of the item to change
 *                NewText: New text for the menu item
 * Return Value - None
 ****************************************************************************/
void AlterMenuText (HMENU hMenu , int ID , char *NewText)
{
 static MENUITEMINFO mii;
 mii.cbSize = sizeof (MENUITEMINFO);

 // Set the MENUITEMINFO struct to inform windows that we want to change the
 // menu item's type and type data.
 mii.fMask = MIIM_TYPE;
 mii.fType = MFT_STRING;
 mii.dwTypeData = NewText;

 // Set the new altered item attributes
 SetMenuItemInfo (hMenu , ID , FALSE , &mii);
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgMenuEnable
 * Purpose      - This handles the WM_INITMENUPOPUP menu. We get this message
 *                before a popup menu is displayed by windows. We handle this
 *                to give us a chance to fiddle with the menu items in the
 *                menu's on all of our windows to check/uncheck and alter
 *                their text as needed.
 * Arguments    - hWnd: Handle of the window that's getting the message
 *                wParam: Handle to the menu being opened
 * Return Value - 1 if we process the message, 0 if not
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgMenuEnable (HWND hWnd , UINT /*uMessage*/ ,
                                           WPARAM wParam , LPARAM /*lParam*/)
{
 // Is the file menu opening? If so do command enabling for it
 if ((HMENU) wParam == GetSubMenu (GetMenu (hWnd) , FILEMENUPOS)) {
     // Alter the text of the menu item for displaying the debug window
     // depending on if it is visible or not.
     if (hwndDebugClient)
         AlterMenuText ((HMENU) wParam , IDM_DISPLAYDEBUG , "Hide &debug window");
     else
         AlterMenuText ((HMENU) wParam , IDM_DISPLAYDEBUG , "Display &debug window");

     return 0;
 }

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::MsgTrayNotify
 * Purpose      - This handles the WM_TRAYNOTIFY message. This is a custom
 *                message of our own devising, and is sent to us whenever
 *                a mouse event occurs on our icon in the tray
 * Arguments    - hWnd: Handle of the window getting the message
 *                lParam: ID of the mouse event that occured
 *                wParam: ID of the tray icon (unused here)
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::MsgTrayNotify (HWND hWnd , UINT /*uMessage*/,
                                           WPARAM /*wParam*/ , LPARAM lParam)
{
 // Determine what mouse event occured so we know what to do.
 switch (lParam) {
     case WM_LBUTTONDBLCLK:
         SetForegroundWindow (hWnd);
         CmdToggleWindow (hWnd , IDM_DISPLAYDEBUG , 0 , 0);
         break;

     case WM_RBUTTONDOWN:
         // Get a handle to the menu that we will be displaying.
         HMENU Menu = GetSubMenu (GetMenu (hWnd) , 0);

         // Set this window to the foreground because of a dumb design
         // decision by microsoft that stops the popup menu from being
         // erased if you click outside it.
         SetForegroundWindow (hWnd);

         // Get the mouse cursor position
         POINT P;
         GetCursorPos (&P);

         // Display the popup menu
         TrackPopupMenu (Menu , TPM_LEFTALIGN , P.x , P.y , 0 , hWnd , NULL);

         // Do another fix so that the next time the menu is displayed it
         // doesn't vanish
         PostMessage (hWnd , WM_NULL , 0 , 0);

         break;
 }

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::CmdToggleWindow
 * Purpose      - This function is called in response to the user choosing the
 *                "Display debug window" command from the menu.
 * Arguments    - hWnd: Window being closed
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::CmdToggleWindow (HWND /*hWnd*/, WORD /*wCommand*/,
                                             WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 // Determine if the user wanted to toggle the debug window and then
 // act accordingly.
 // If there is a debug client already, then send it a close message, which
 // will cause a destroy message, which will remove the window from the
 // screen. Otherwise, create the window.
 if (hwndDebugClient) {
     PostMessage (hwndDebugClient , WM_CLOSE , 0 , 0);
 } else {
     // Create the window
     CResString Title (hInst , IDS_DEBUGCAPTION);

     hwndDebugClient = CreateWindowEx (WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
                                       szAppName,
                                       Title,
                                       WS_OVERLAPPEDWINDOW,
                                       CW_USEDEFAULT, 0,
                                       CW_USEDEFAULT, 0,
                                       NULL,
                                       NULL,
                                       hInst,
                                       NULL);

     // Now that the window is created, shift the debug edit window
     // ownership to the window.
     SetParent (hwndDebugEdit , hwndDebugClient);

     // Attempt to load the saved position from the disk
     WINDOWPLACEMENT wp;
     if (LoadDebugWindowPostion (wp)) {
         // The position existed and we loaded it in. Set the window placement
         // using that information. Before we do that, we make sure that it will
         // actually be visible on the screen.

         // Get the dimensions of the screen.
         int SX = GetSystemMetrics (SM_CXSCREEN);
         int SY = GetSystemMetrics (SM_CYSCREEN);

         // First, make sure the minimized position of the window is within the
         // screen bounds.
         if (wp.ptMinPosition.x > SX)
             wp.ptMinPosition.x = SX / 2;
         if (wp.ptMinPosition.y > SY)
             wp.ptMinPosition.y = SY / 2;

         // Next, make sure the maximized position of the window is within screen
         // bounds.
         if (wp.ptMaxPosition.x > SX)
             wp.ptMaxPosition.x = SX / 2;
         if (wp.ptMaxPosition.y > SY)
             wp.ptMaxPosition.y = SY / 2;

         // Finally, make sure that the normal postion of the window falls withiin
         // screen bounds.
         if (wp.rcNormalPosition.left > SX)
             wp.rcNormalPosition.left = SX / 2;
         if (wp.rcNormalPosition.right > SX)
             wp.rcNormalPosition.right = SX;
         if (wp.rcNormalPosition.top > SY)
             wp.rcNormalPosition.top = SY / 2;
         if (wp.rcNormalPosition.bottom > SY)
             wp.rcNormalPosition.bottom = SY;

         // Now, set the window's placement
         SetWindowPlacement (hwndDebugClient , &wp);
     } else
         ShowWindow (hwndDebugClient , SW_SHOW);

     // Update the new window
     UpdateWindow (hwndDebugClient);
 }

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::CmdClearDebug
 * Purpose      - This function is called in response to the user choosing the
 *                clear debug menu item from the popup menu.
 * Arguments    - hWnd: Window being closed
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::CmdClearDebug (HWND /*hWnd*/, WORD /*wCommand*/,
                                           WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 // Select everything that's in the window.
 CHARRANGE cr;
 cr.cpMin = 0;
 cr.cpMax = -1;
 SendMessage (hwndDebugEdit , EM_EXSETSEL , 0 , (LPARAM) &cr);

 // Now tell it to clear out the selected text
 SendMessage (hwndDebugEdit , EM_REPLACESEL , FALSE , (LPARAM) "");

 return 0;
}

/*****************************************************************************
 * Name         - CMainWndDispatcher::CmdExit
 * Purpose      - This function is called in response to the user choosing the
 *                exit menu item from the popup menu. It simply tells the
 *                window to close.
 * Arguments    - hWnd: Window being closed
 * Return Value - 0
 ****************************************************************************/
LRESULT CMainWndDispatcher::CmdExit (HWND /*hWnd*/, WORD /*wCommand*/,
                                     WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 PostMessage (hwndMain , WM_CLOSE , 0 , 0);
 return 0;
}

/*****************************************************************************
 * Name         - InitializeMainWindow
 * Purpose      - This sets up the main window by creating the two edit
 *                controls used in the debug and note windows and assigns
 *                them the main window as their parent. The main window is
 *                created hidden and forced hidden so they don't immediatly
 *                appear. In addition the tray icon is set up.
 * Arguments    - Parent: Handle to the main window
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL InitializeMainWindow (HWND Parent)
{
 // Try to create the debug and note edit windows. If either fails we return
 // -1 to stop program execution.
 hwndDebugEdit = CreateWindow ("richedit" , "" ,
                               WS_VSCROLL | WS_VISIBLE | ES_READONLY |
                               ES_MULTILINE | ES_AUTOVSCROLL | WS_CHILD | ES_SUNKEN |
                               ES_DISABLENOSCROLL ,
                               0 , 0 , 0 , 0 , Parent ,
                               (HMENU) 0 ,
                               hInst ,
                               NULL);
 if (hwndDebugEdit == NULL)
     return FALSE;

 // Get a slimmer font for the debug text and tell the window to use it.
 HFONT Font = (HFONT) GetStockObject (ANSI_VAR_FONT);
 SendMessage (hwndDebugEdit , WM_SETFONT , (WPARAM) Font , TRUE);

 // Load up the icon that will be used in the tray
 HICON TrayIcon = (HICON) LoadImage (hInst,
                                     MAKEINTRESOURCE (IDI_MAINICON),
                                     IMAGE_ICON,
                                     16 , 16 , 0);
 if (TrayIcon == NULL)
     TrayIcon = LoadIcon (NULL , IDI_APPLICATION);

 // Load the string that will be used for the tooltip on the icon
 CResString Tooltip (hInst , IDS_TIPTEXT);

 // Set up the structure that tells windows how to set up the tray icon
 NOTIFYICONDATA nid;

 nid.cbSize = sizeof (NOTIFYICONDATA);
 nid.hWnd = Parent;
 nid.uID = 0;
 nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
 nid.uCallbackMessage = WM_TRAYNOTIFY;
 nid.hIcon = TrayIcon;
 strncpy (nid.szTip , Tooltip , 63);
 nid.szTip[63] = '\0';

 // Try to add the icon. If it fails return false from the create message
 // to stop the application, because the whole program interface is through
 // the tray.
 if (Shell_NotifyIcon (NIM_ADD , &nid) == 0)
     return FALSE;

 // If we get here it everything worked out OK
 return TRUE;
}

/****************************************************************************/


